// (c) COPYRIGHT URI 2000
// Please read the full copyright statement in the file COPYRIGHT.
//
// Authors:
//      Reza            Reza Nekovei <rnekovei@ieee.org>
//      jhrg,jimg       James Gallagher <jgallagher@gso.uri.edu>

// Implementation for Translation.
//

#ifdef __GNUG__
#pragma implementation
#endif

#include "config_dap.h"

#include <string>

#include "Pix.h"

#include "util.h"
#include "Structure.h"
#include "Sequence.h"
#include "Array.h"
#include "Grid.h"
#include "DDS.h"
#include "Translation.h"
#include "Connect.h"
#include "cgi_util.h"

Translation::Translation(const DDS &rdds, const string &at) 
  : tdds(rdds), dds(rdds), accept_types(at)
{
}

DDS& Translation::get_tdds()
{
  TransType();
  return tdds;
}

bool
Translation::send_servers_data()
{
  return true;
}

void
Translation::copy_data(BaseType *src, BaseType *dest)
{
}

void
Translation::print_variable_mapping(ostream &os)
{
  for (vector<bt_pair>::size_type i = 0; i < variable_mapping.size(); ++i)
      variable_mapping[i].print(os);
}

BaseType *
Translation::find_translated_var(BaseType *btp)
{
    return 0;
}

void 
Translation::write_binary_data(const string &dataset, const string &cgi_ver, 
			       FILE *out)
{
    set_mime_binary(out, dods_data, cgi_ver, x_plain);
   
    tdds.print_constrained(out);

    fprintf(out, "Data:\n");

    XDR *xdr_sink = new_xdrstdio(out, XDR_ENCODE);
    
    bool status = true;
    for (Pix p = dds.first_var(); p && status; dds.next_var(p)) {
	if (dds.var(p)->send_p()) { // only process projected variables
	    status = dds.var(p)->serialize(dataset, dds, xdr_sink, true);
	}
    }

#if 0
    dds.write_binary_data(dataset, xdr_sink);
#endif
}

// find unsupported constructor types and 
// translate them to the acceptable types
// structures using a translation lookup table
// List, Function, and URL types are not 
// currently supported.
void
Translation::TransType()
{
    int TransTable = get_trans_table(); 

    switch (TransTable & 0x07) {
      case 0:			// Nothing to translate to
	remover();
	break;
      case 1: 
	grid2seq();
      case 3:
	array2seq();
	break;
      case 2:			// Can not grid gappy data
	nogappygrid();
	break;
      case 4:
	seq2array();
      case 5:
	grid2array();
	break;
      case 6:
	seq2array();
	break;
      case 7:
	break ;			// 0;
    }      

    if(TransTable < 8)		// Use variable_rename to replace structures
	flatstruc();
}

void 
Translation::add_variable_mapping(BaseType *orig, BaseType *trans)
{
    variable_mapping.push_back(bt_pair(orig, trans));
}

/////////////////////////////////////////////////////
// The following DDS convertors would propagate through all levels of
// structures 
//  1) remover(), remove grids, sequences and arrays from DDS 
//  2) grid2seq()
//  3) array2seq()
//  4) grid2array()
//  5) seq2array()
//  6) flatstruc(), variable rename
//  7) nogappygrid(), remove sequence and array

const string spr = "."; 

// Remove all the non-supported types 
void 
Translation::remover() 
//make it recursive to clean inside structures as well <--
{
    string var_name;

    for(Pix Cp=tdds.first_var(); Cp;){

	if((tdds.var(Cp)->type() == dods_array_c ) ||
	   (tdds.var(Cp)->type() == dods_sequence_c ) || 
	   (tdds.var(Cp)->type() == dods_grid_c )){ 

	    // If the variable has been deleted, there's no mapping.
	    add_variable_mapping(tdds.var(Cp), 0);
	    var_name = tdds.var(Cp)->name();

	    tdds.next_var(Cp);
	    tdds.del_var(var_name);
	}
	else
	    tdds.next_var(Cp);
    }
}

void 
Translation::flatstruc()
{
    bool nested = false;
    string var_name;
  
    for (Pix Sp=tdds.first_var(); Sp;) {

	if (tdds.var(Sp)->type() == dods_structure_c) {
	    add_variable_mapping(tdds.var(Sp), 0);

	    var_name = tdds.var(Sp)->name();
	    Structure *stq = (Structure *) tdds.var(Sp);

	    for (Pix p=stq->first_var(); p!=0; stq->next_var(p)) {

		if (stq->var(p)->type() == dods_structure_c)
		    nested = true;	  
		//  stq->var(p)->set_name(var_name + "." + stq->var(p)->name());   	
		BaseType *btp = stq->var(p)->ptr_duplicate();

		if(btp->type() == dods_array_c)
		    btp->var()->set_name(var_name + spr + stq->var(p)->name()); 	
		// else
		btp->set_name(var_name + spr + stq->var(p)->name());	
		tdds.add_var(btp);
		add_variable_mapping(stq->var(p), btp);

		// tdds.add_var(stq->var(p));
	    }
	    tdds.next_var(Sp);	
	    tdds.del_var(var_name);  
	}
	else
	    tdds.next_var(Sp);
    } 

    if (nested)
	flatstruc();
}
void 
Translation::nogappygrid()
{
    // Sequences are removed since their translation is lossy
    // and arrays are indexed to create Grids.<--
    string var_name;
  
    for(Pix Cp=tdds.first_var(); Cp;){

	if((tdds.var(Cp)->type() == dods_sequence_c)||
	   (tdds.var(Cp)->type() == dods_array_c)){ 

	    var_name = tdds.var(Cp)->name();

	    add_variable_mapping(tdds.var(Cp), 0);
	    tdds.next_var(Cp);
	    tdds.del_var(var_name);
	}
	else
	    tdds.next_var(Cp);
    }
}

void 
Translation::grid2seq()
{//Multiple indeces can show the relation between grid and its map vectors
    for(Pix Cp=tdds.first_var(); Cp;){

	if(tdds.var(Cp)->type() == dods_grid_c ){

	    string var_name = tdds.var(Cp)->name();
	    Sequence *seq = NewSequence(var_name);
	    Grid *gr = (Grid *) tdds.var(Cp);
	    BaseType *arbtp = gr->array_var()->var()->ptr_duplicate();
	    seq->add_var(arbtp);
	    add_variable_mapping(gr->array_var()->var(), arbtp);

	    for(Pix q = gr->first_map_var(); q; gr->next_map_var(q)){
		BaseType *mpbtp = gr->map_var(q)->var()->ptr_duplicate();
		seq->add_var(mpbtp);
		add_variable_mapping(gr->map_var(q)->var(), mpbtp);
	    }

	    add_variable_mapping(tdds.var(Cp), seq);
	    tdds.next_var(Cp);
	    tdds.del_var(var_name);
	    tdds.add_var(seq);
	}
	else
	    tdds.next_var(Cp);
    }
}

void 
Translation::array2seq()
{// An index would help to show the element order <--

    string var_name;

    for(Pix Cp=tdds.first_var(); Cp;){

	if(tdds.var(Cp)->type() == dods_array_c ){

	    var_name = tdds.var(Cp)->name();
	    Sequence *seq = NewSequence(var_name);
	    BaseType *btp = tdds.var(Cp)->var()->ptr_duplicate();
	    seq->add_var(btp);
	    add_variable_mapping(tdds.var(Cp)->var(), btp);

	    add_variable_mapping(tdds.var(Cp), seq);
	    tdds.next_var(Cp);
	    tdds.del_var(var_name);
	    tdds.add_var(seq);
	}
	else
	    tdds.next_var(Cp);
    }
}
void 
Translation::grid2array()
{ 
    string var_name;

    for(Pix Cp=tdds.first_var(); Cp;){

	if(tdds.var(Cp)->type() == dods_grid_c ){
	    var_name = tdds.var(Cp)->name();
	    Structure *stq = NewStructure(var_name);
	    Grid *gr = (Grid *) tdds.var(Cp);
	    BaseType *arbtp = gr->array_var()->ptr_duplicate();
	    stq->add_var(arbtp);
	    add_variable_mapping(gr->array_var(), arbtp);

	    //stq->add_var(gr->array_var());

	    for(Pix q = gr->first_map_var(); q; gr->next_map_var(q)){
		BaseType *mpbtp = gr->map_var(q)->ptr_duplicate();
		stq->add_var(mpbtp);
		add_variable_mapping(gr->map_var(q), mpbtp);
		//stq->add_var(gr->map_var(q));
	    }

	    add_variable_mapping(tdds.var(Cp), stq);
	    tdds.next_var(Cp);
	    tdds.del_var(var_name);
	    tdds.add_var(stq);
	}
	else
	    tdds.next_var(Cp);
    }
}

void 
Translation::seq2array()
{
    bool nested = false;
    string var_name;
    int SeqRec = 100; //dummy test number

    for (Pix Sp=tdds.first_var(); Sp;){

	if (tdds.var(Sp)->type() == dods_sequence_c){
	    var_name = tdds.var(Sp)->name();

	    Sequence *seq = (Sequence *) tdds.var(Sp);
	    // SeqRec = seq->length(); 
	    //if (SeqRec > 0){ // can translate otherwise remove the sequence

	    for (Pix p=seq->first_var(); p!=0; seq->next_var(p)){
		BaseType *btp = seq->var(p)->ptr_duplicate();

		if (btp->type() == dods_array_c)
		    btp->var()->set_name(var_name + spr + seq->var(p)->name()); 	
		// else
		btp->set_name(var_name + spr + seq->var(p)->name());	

		if (btp->type() == dods_sequence_c) {
		    nested = true;
		    tdds.add_var(btp);
		    add_variable_mapping(seq->var(p), btp);
		} 
		else if (btp->type() == dods_array_c){
		    Array *ar = (Array *) btp;
		    ar->append_dim(SeqRec,"record");
		    tdds.add_var(ar);
		    add_variable_mapping(seq->var(p), ar);
		}
		else {
		    Array *ar = NewArray(var_name);
		    ar->add_var(btp);
		    ar->append_dim(SeqRec,"record");

		    tdds.add_var(ar);
		    add_variable_mapping(seq->var(p), ar);
		}
	
	    }
	    // } // sequence length check
	    add_variable_mapping(tdds.var(Sp), 0);
	    tdds.next_var(Sp);	
	    tdds.del_var(var_name);  
	}
	else
	    tdds.next_var(Sp);
    } 
    if (nested)
	seq2array();
}


int
Translation::get_trans_table()
{
    int trans_table;
    bool structure;

    for(unsigned int i=0; i<accept_types.length(); i++)
	accept_types[i] = tolower(accept_types[i]);

    if (accept_types.find("all") != string::npos) 
	return (7); // No need for translations
      
    // parse acceptable types
    if (accept_types.find("!") == string::npos) { 
	// Initialize positive selection
	trans_table = 0;
	structure = false;

	if (accept_types.find("sequence") != string::npos)
	    trans_table = 1;
	if (accept_types.find("grid") != string::npos)
	    trans_table = trans_table + 2;
	if (accept_types.find("array") != string::npos)
	    trans_table = trans_table + 4;
	if (accept_types.find("structure") != string::npos)
	    structure = true;
    }
    else { 
	// Initialize negative selection
	trans_table = 7;
	structure = true;

	if (accept_types.find("!sequence") != string::npos)
	    trans_table = 6;
	if (accept_types.find("!grid") != string::npos)
	    trans_table = trans_table - 2;
	if (accept_types.find("!array") != string::npos)
	    trans_table = trans_table - 4;
	if (accept_types.find("!structure") != string::npos)
	    structure = false;
    }
    if(structure) trans_table=8+trans_table;
    return trans_table;
}

